import abc

from basics.base import Base


class ConversationEngineBase(Base, metaclass=abc.ABCMeta):

    def __init__(self, io_processor, model_evaluator, max_sequence_length=40, debug=False):
        super().__init__()

        self._io_processor = io_processor
        self._model_evaluator = model_evaluator

        self._max_sequence_length = max_sequence_length

        self._debug = debug

        self._conversation_context = {}

        self._validate()

    def reset_state(self):
        self._conversation_context = {}
        self._model_evaluator.reset_state()

    def execute_command(self, command, user_name=None):
        """

        :param command:
        :param user_name: Optional, user name of the user who input the command

        :return: <system message>, <bot response>
                 = None, None when the conversation engine did not process any command
        """
        return None, None

    @abc.abstractmethod
    def respond(self, inputs, conversation_start=False):
        """

        :param inputs: Dict with one or more different toes of inputs
        :param conversation_start: Boolean

        :return: <response to input>, <score(s)>
        """
        self._log.error("Please implement this method in a child class")

    def _validate(self):
        # TODO
        self._valid = True


class BasicConversationEngine(ConversationEngineBase):

    def __init__(self,
                 io_processor,
                 model_evaluator,
                 select_token_func,
                 sequence_end_index=None,
                 is_sequence_end_func=None,
                 **kwargs):
        super().__init__(io_processor, model_evaluator, **kwargs)

        self._select_token_func = select_token_func

        if (sequence_end_index is None and is_sequence_end_func is None) or \
                (sequence_end_index is not None and is_sequence_end_func is not None):
            raise Exception("Either define sequence_end_index or is_sequence_end_func")

        if sequence_end_index is not None:
            self._log.debug("Creating is_sequence_end_func using sequence_end_index")
            is_sequence_end_func = lambda response: response[-1] == sequence_end_index

        self._sequence_end_index = sequence_end_index
        self._is_sequence_end_func = is_sequence_end_func

    def respond(self, inputs, conversation_start=False):
        processed_inputs = self._io_processor.process_inputs(inputs, conversation_start)

        self._model_evaluator.update_context(processed_inputs, self._conversation_context, conversation_start)

        return self._create_response()

    def _create_response(self):
        """
        Called after model context updated

        :return:
        :rtype:
        """

        prediction_context = {}
        prev_token = None
        response = []
        scores = []
        for _ in range(self._max_sequence_length):

            prediction_data = self._model_evaluator.predict_next_token(prev_token,
                                                                       prediction_context,
                                                                       self._conversation_context)
            # Obtain most likely word token and its score
            score, token = self._select_token_func(prediction_data)

            # Record token and score
            response += [self._unwrap(token)]
            scores += [self._unwrap(score)]

            if self._is_sequence_end_func(response):
                break

            prev_token = token

        response, scores = self._io_processor.process_response(response, scores)

        return response, scores

    def _unwrap(self, tensor):
        """
        Unwrap scalar tensor

        :param tensor:
        :return:
        """
        return tensor



import abc

from basics.base import Base


class IOProcessorBase(Base, metaclass=abc.ABCMeta):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def process_inputs(self, inputs, conversation_start):
        """

        :param inputs: Dict with one or more different types of inputs
        :param conversation_start: Bool, if the given inputs are the initial inputs of a new
                                   conversation

        :return:
        """
        self._log.error("Please implement this method in a child class")

    @abc.abstractmethod
    def process_response(self, response, scores):
        """

        :param response:
        :param scores:

        :return: processed response, scores
        """
        self._log.error("Please implement this method in a child class")


class SimpleIOProcessor(IOProcessorBase):

    def __init__(self,
                 max_sequence_length,
                 preprocessing_func,
                 indexing_func,
                 deindexing_func,
                 **kwargs):
        """

        :param max_sequence_length : Max. sequence length of any input sequences

        :param preprocessing_func  : function that takes a list of sequences as inputs and
                                     returns a preprocessed list of inputs

        :param indexing_func       : function that takes a list of sequences as inputs
                                     (not tokenized) and returns a tokenized and indexed list of inputs

        :param deindexing_func     : function that takes a list of tokenized and indexed inputs
                                     and returns a detokenized and deindexed list of inputs
        """
        super().__init__(**kwargs)

        self._max_sequence_length = max_sequence_length
        self._preprocessing_func = preprocessing_func
        self._indexing_func = indexing_func
        self._deindexing_func = deindexing_func

    def process_inputs(self, inputs, conversation_start):
        """

        :param inputs: Dict with one or more different types of inputs
        :param conversation_start: Bool, if the given inputs are the initial inputs of a new
                                   conversation

        :return:
        """
        inputs = self._get_input_chats(inputs, conversation_start)
        inputs = self._preprocessing_func(inputs)
        inputs = self._indexing_func(inputs)
        inputs = self._cut_off_too_long(inputs)
        inputs = self._build_input_sequences(inputs)

        return inputs

    def process_response(self, response, scores):
        """

        :param inputs: Dict with one or more different types of inputs
        :return:
        """

        # deindexing_func assumes a list of sequences
        return self._deindexing_func([response])[0], scores

    @abc.abstractmethod
    def _get_input_chats(self, inputs, conversation_start):
        """

        Get the raw input chats (and other inputs if any) from the inputs dict

        :param inputs: Dict with one or more different types of inputs
        :param conversation_start: Bool, if the given inputs are the initial inputs of a new
                                   conversation

        :return:
        """
        raise NotImplementedError("Please implement this method in a child class")

    def _cut_off_too_long(self, inputs):
        inputs_checked = []
        for input in inputs:
            if len(input) > self._max_sequence_length-1:
                input = input[:self._max_sequence_length-1]

            inputs_checked += [input]

        return inputs_checked

    def _build_input_sequences(self, inputs):
        raise NotImplementedError("Please implement this method in a child class")


from botshop.pytorch.conversation_engine import BasicConversationEngine
from botshop.sample_and_rank import SampleAndRank


class SampleAndRankConversationEngine(SampleAndRank, BasicConversationEngine):
    pass



import torch
import torch.nn.functional as F


def random_sample(logits, T):
    p = F.softmax(logits / T, dim=0)

    token_idx = torch.multinomial(p.view(-1), 1)
    token_p = torch.gather(p, 0, token_idx)

    return token_p, token_idx


def select_max(logits):
    p = F.softmax(logits, dim=0)

    token_p, token_idx = torch.max(p, dim=0)

    return token_p, token_idx


def top_p_sample(p, top_p=0.9, temperature=1.0):
    """

    TODO : make it work for batch_size > 0

    :param p: shape (1, batch_size, num_tokens)
    :param top_p:
    :param temperature:
    :return:
    """

    p_sorted, sorted_indices = torch.sort(p, dim=2, descending=True)
    p_cumulative = torch.cumsum(p_sorted, dim=2)

    to_keep = p_cumulative <= top_p

    # Always keep at least one, so shift to the right
    to_keep[..., 1:] = to_keep[..., :-1].clone()
    to_keep[..., 0] = 1

    to_keep = sorted_indices[to_keep]

    p = p[..., to_keep]

    token_p, token_idx = sample(p, temperature)

    token_idx = to_keep[..., token_idx]

    return token_p, token_idx


def sample(p, temperature=1.0):
    """
    TODO : make it work for batch_size > 0

    :param p: shape (1, 1, num_tokens)
    :param temperature:
    :return:
    """
    p = torch.pow(p, 1.0 / temperature)
    p_sum = torch.sum(p, dim=2)
    p = p / p_sum[:, :, None]

    token_idx = torch.multinomial(p.view(-1), 1)
    token_p = torch.gather(p, 2, token_idx.view(1, -1, 1))

    token_idx = token_idx.view(1, -1)
    token_p = token_p.view(1, -1)

    return token_p, token_idx



import re
import statistics

from basics.base import Base

from basics.logging_utils import log_exception


class SimpleBot(Base):
    SYSTEM_COMMAND = re.compile('^##(.+)##$')

    SET_INITIAL_RESPONSE_COMMAND = re.compile('^INIT_RESPONSE="(.*)"$', re.I)
    NO_INITIAL_RESPONSE_COMMAND = re.compile('^NO_INIT_RESPONSE$', re.I)

    RESET_STATE_COMMAND = re.compile('^(reset|r)$', re.I)

    QUIT_COMMAND = re.compile('^(q|quit)$', re.I)

    def __init__(self,
                 conversation_engine,
                 user_name='User',
                 bot_name='Bot',
                 init_response=None,
                 debug=False):
        """

        :param conversation_engine: Conversation engine instance derived from ConversationEngineBase
        :param user_name: Optional, default user name if no user name given with respond_to() method
        :param bot_name: Optional, default bot name
        :param init_response: Optional, initial bot response

        :param debug: If True, more logging is done
        """

        super().__init__()

        self._conversation_engine = conversation_engine

        self._user_name = user_name
        self._bot_name = bot_name

        self._debug = debug

        # chat history
        self._chats = []
        self._is_user = []
        self._actor_name = []

        self._conversation_start = True

        self._init_response = None
        self.set_initial_response(init_response)

        self.reset_state()

    def reset_state(self):
        if self._debug:
            self._log.debug("Resetting state ...")

        self._conversation_engine.reset_state()

        self._chats = []
        self._is_user = []
        self._actor_name = []

        self._conversation_state = None
        self._conversation_start = True
        if self._init_response:
            if self._debug:
                self._log.debug('Adding initial response: %s' % self._init_response)

            self._chats += [self._init_response]
            self._is_user += [False]
            self._actor_name += [self._bot_name]

            return self._init_response

    def set_initial_response(self, r):
        self._init_response = r

        if self._debug:
            self._log.debug('Set initial response : %s' % self._init_response)

    def respond_to(self, user_chat, user_name=None):
        if user_name is None:
            user_name = self._user_name

        if self._debug:
            self._log.debug(f'{user_name} : {user_chat}')

        scores = None
        system_message, bot_chat = self._execute_command_in(user_chat, user_name=user_name)
        if system_message is None:  # No command executed
            self._chats.append(user_chat)
            self._is_user.append(True)
            self._actor_name.append(user_name)

            bot_chat, scores = self._respond()

            if self._conversation_start:
                self._conversation_start = False

            self._chats.append(bot_chat)
            self._is_user.append(False)
            self._actor_name.append(self._bot_name)
        else:
            self._log.info('SYSTEM : \n%s' % system_message)

        if bot_chat:
            self._log.info('%s [%f] : \n%s' % (self._bot_name, statistics.mean(scores), bot_chat))

        return {
            'bot': bot_chat,
            'scores': scores,
            'system': system_message,
        }

    def _execute_command_in(self, user_chat, user_name=None):
        if user_name is None:
            user_name = self._user_name

        m = self.SYSTEM_COMMAND.match(user_chat)

        if m is None:
            return None, None

        try:
            command = m.groups()[0]

            m = self.QUIT_COMMAND.match(command)
            if m is not None:
                system_msg = "quit"
                return system_msg, None

            m = self.RESET_STATE_COMMAND.match(command)
            if m is not None:
                init_bot_response = self.reset_state()
                system_msg = "Conversation model state has been reset."
                return system_msg, init_bot_response

            m = self.SET_INITIAL_RESPONSE_COMMAND.match(command)
            if m is not None:
                try:
                    self.set_initial_response(m.groups()[0])
                except Exception as e:
                    log_exception(self._log, 'An exception occurred parsing the initial response', e)
                    return 'Unable to parse new initial response, initial response not set.', None

                return "New initial response : %s" % (self._init_response if self._init_response else
                                                      "No initial response will be used"), None

            m = self.NO_INITIAL_RESPONSE_COMMAND.match(user_chat)
            if m is not None:
                self.set_initial_response(None)
                return "New initial response : %s" % (self._init_response if self._init_response else
                                                      "No initial response will be used"), None

            return self._forward_command_to_conversation_engine(command, user_name=user_name)
        except Exception as e:
            log_exception(self._log, 'An exception occurred parsing the system command', e)
            return 'Unable to parse system command, no effect.', None

    def _forward_command_to_conversation_engine(self, command, user_name=None):
        if user_name is None:
            user_name = self._user_name

        try:
            return self._conversation_engine.execute_command(command, user_name=user_name)
        except Exception as e:
            log_exception(self._log, 'Forwarding system command to evaluator failed', e)
            return 'Execution of system command by evaluator failed, no effect.', None

    def _respond(self):
        response, scores = self._conversation_engine.respond({
            "chats": self._chats,
            "is_user": self._is_user,
            "actor_name": self._actor_name
        }, self._conversation_start)

        return response, scores




import abc

from basics.base import Base


class ModelEvaluatorBase(Base, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def reset_state(self):
        self._log.error("Please implement this method in a child class")

    @abc.abstractmethod
    def update_context(self, inputs, conversation_context, conversation_start):
        """

        :param inputs: Dict with one or more different types of inputs
        :param conversation_context: Dict, can be empty initially, and can be filled with
                                     context data when calling this method
        :param conversation_start: Boolean

        :return:
        """
        self._log.error("Please implement this method in a child class")

    @abc.abstractmethod
    def predict_next_token(self, previous_token, prediction_context, conversation_context):
        """

        :param previous_token: Previous token, initially None
        :param prediction_context: Dict, can be used to store intermediate state
        :param conversation_context: Dict, conversation context, updated after each turn using `update_context`

        :return:
        """
        self._log.error("Please implement this method in a child class")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="botshop",
    version="0.0.11",
    author="Freddy Snijder",
    author_email="botshop@visionscapers.com",
    description="Framework for chatbot model inference, optimized for neural conversation models.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nuhame/botshop",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'visionscaper-pybase'
    ],
    dependency_links=[],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)